""" built-in PEP filters """
from typing import Dict


def basic_pep_filter(p, **kwargs) -> Dict[str, str]:
    """
    Basic PEP filter, that does not convert the Project object.

    This filter can save the PEP representation to file, if kwargs include `path`.

    :param peppy.Project p: a Project to run filter on
    """
    return {"project": str(p)}


def yaml_samples_pep_filter(p, **kwargs) -> Dict[str, str]:
    """
    YAML samples PEP filter, that returns only Sample object representations.

    This filter can save the YAML to file, if kwargs include `path`.

    :param peppy.Project p: a Project to run filter on
    """
    from yaml import dump

    samples_yaml = []
    for s in p.samples:
        samples_yaml.append(s.to_dict())

    return {"samples": dump(samples_yaml, default_flow_style=False)}


def yaml_pep_filter(p, **kwargs) -> Dict[str, str]:
    """
    YAML PEP filter, that returns Project object representation.

    This filter can save the YAML to file, if kwargs include `path`.

    :param peppy.Project p: a Project to run filter on
    """
    from yaml import dump

    data = p.config.to_dict()
    return {"project": dump(data, default_flow_style=False)}


def csv_pep_filter(p, **kwargs) -> Dict[str, str]:
    """
    CSV PEP filter, that returns Sample object representations

    This filter can save the CSVs to files, if kwargs include
    `sample_table_path` and/or `subsample_table_path`.

    :param peppy.Project p: a Project to run filter on
    """
    sample_table_path = kwargs.get("sample_table_path")
    subsample_table_path = kwargs.get("subsample_table_path")
    sample_table_repr = p.sample_table.to_csv(path_or_buf=sample_table_path)

    s = ""
    if sample_table_repr is not None:
        s += sample_table_repr
    if p.subsample_table is not None:
        subsample_table_repr = p.subsample_table.to_csv(
            path_or_buf=subsample_table_path
        )
        if subsample_table_repr is not None:
            s += subsample_table_repr

    return {"samples": s}


def processed_pep_filter(p, **kwargs) -> Dict[str, str]:
    """
    Processed PEP filter, that returns the converted sample and subsample tables.
    This filter can return the tables as a table or a document.
    :param peppy.Project p: a Project to run filter on
    :param bool samples_as_objects: Flag to write as a table
    :param bool subsamples_as_objects: Flag to write as a table
    """
    # get params
    samples_as_objects = kwargs.get("samples_as_objects")
    subsamples_as_objects = kwargs.get("subsamples_as_objects")

    prj_repr = p.config.to_dict()

    return {
        "project": str(prj_repr),
        "samples": str(p.samples)
        if samples_as_objects
        else str(p.sample_table.to_csv()),
        "subsamples": str(p.subsamples)
        if subsamples_as_objects
        else str(p.subsample_table.to_csv()),
    }



"""
Project configuration
"""

from ._version import __version__
from .const import *
from .conversion import *
from .conversion_plugins import *
from .exceptions import *
from .inspection import *
from .schema import *
from .validation import *

__all__ = [
    "validate_project",
    "validate_sample",
    "validate_config",
    "read_schema",
    "inspect_project",
    "get_available_pep_filters",
    "convert_project",
    "basic_pep_filter",
    "yaml_pep_filter",
    "csv_pep_filter",
    "yaml_samples_pep_filter",
]



import inspect
import sys
import os
from logging import getLogger

from pkg_resources import iter_entry_points

from .exceptions import *

_LOGGER = getLogger(__name__)


def pep_conversion_plugins():
    """
    Plugins registered by entry points in the current Python env

    :return dict[dict[function(peppy.Project)]]: dict which keys
        are names of all possible hooks and values are dicts mapping
        registered functions names to their values
    :raise EidoFilterError: if any of the filters has an invalid signature.
    """
    plugins = {}
    for ep in iter_entry_points("pep.filters"):
        plugin_fun = ep.load()
        if len(list(inspect.signature(plugin_fun).parameters)) != 2:
            raise EidoFilterError(
                f"Invalid filter plugin signature: {ep.name}. "
                f"Filter functions must take 2 arguments: peppy.Project and **kwargs"
            )
        plugins[ep.name] = plugin_fun
    return plugins


def convert_project(prj, target_format, plugin_kwargs=None):
    """
    Convert a `peppy.Project` object to a selected format

    :param peppy.Project prj: a Project object to convert
    :param dict plugin_kwargs: kwargs to pass to the plugin function
    :param str target_format: the format to convert the Project object to
    :raise EidoFilterError: if the requested filter is not defined
    """
    return run_filter(prj, target_format, plugin_kwargs or dict())


def run_filter(prj, filter_name, verbose=True, plugin_kwargs=None):
    """
    Run a selected filter on a peppy.Project object

    :param peppy.Project prj: a Project to run filter on
    :param str filter_name: name of the filter to run
    :param dict plugin_kwargs: kwargs to pass to the plugin function
    :raise EidoFilterError: if the requested filter is not defined
    """
    # convert to empty dictionary if no plugin_kwargs are passed
    plugin_kwargs = plugin_kwargs or dict()

    # get necessary objects
    installed_plugins = pep_conversion_plugins()
    installed_plugin_names = list(installed_plugins.keys())
    paths = plugin_kwargs.get("paths")
    env = plugin_kwargs.get("env")

    # set environment
    if env is not None:
        for var in env:
            os.environ[var] = env[var]

    # check for valid filter
    if filter_name not in installed_plugin_names:
        raise EidoFilterError(
            f"Requested filter ({filter_name}) not found. "
            f"Available: {', '.join(installed_plugin_names)}"
        )
    _LOGGER.info(f"Running plugin {filter_name}")
    func = installed_plugins[filter_name]

    # run filter
    conv_result = func(prj, **plugin_kwargs)

    # if paths supplied, write to disk
    if paths is not None:
        # map conversion result to the
        # specified path
        for result_key in conv_result:
            result_path = paths.get(result_key)
            if result_path is None:
                _LOGGER.warning(
                    f"Conversion plugin returned key that doesn't exist in specified paths: '{result_key}'."
                )
            else:
                # create path if it doesn't exist
                if not os.path.exists(result_path):
                    os.makedirs(os.path.dirname(result_path), exist_ok=True)
                # write to path
                with open(result_path, "w") as f:
                    f.write(conv_result[result_key])

    if verbose:
        for result_key in conv_result:
            sys.stdout.write(conv_result[result_key])
    else:
        # simply return from the function with
        # conversion results
        pass

    return conv_result


def get_available_pep_filters():
    """
    Get a list of available target formats

    :return List[str]: a list of available formats
    """
    return list(pep_conversion_plugins().keys())



def inspect_project(p, sample_names=None, max_attr=10):
    """
    Print inspection info: Project or,
    if sample_names argument is provided, matched samples

    :param peppy.Project p: project to inspect
    :param Iterable[str] sample_names: list of samples to inspect
    :param int max_attr: max number of sample attributes to display
    """
    if sample_names:
        samples = p.get_samples(sample_names)
        if not samples:
            print("No samples matched by names: {}".format(sample_names))
            return
        for s in samples:
            print(s.__str__(max_attr=max_attr))
            print("\n")
        return
    print(p)
    return



import os
from copy import deepcopy as dpcpy
from logging import getLogger
from warnings import catch_warnings as cw
from .exceptions import EidoValidationError

from pandas.core.common import flatten
from ubiquerg import size

from jsonschema import Draft7Validator

from .const import (
    ALL_INPUTS_KEY,
    FILES_KEY,
    INPUT_FILE_SIZE_KEY,
    MISSING_KEY,
    PROP_KEY,
    REQUIRED_FILES_KEY,
    REQUIRED_INPUTS_KEY,
)
from .schema import preprocess_schema, read_schema

_LOGGER = getLogger(__name__)


def _validate_object(object, schema, exclude_case=False):
    """
    Generic function to validate object against a schema

    :param Mapping object: an object to validate
    :param str | dict schema: schema dict to validate against or a path to one
    :param bool exclude_case: whether to exclude validated objects
        from the error. Useful when used ith large projects
    """

    validator = Draft7Validator(schema)
    if not validator.is_valid(object):
        errors = sorted(validator.iter_errors(object), key=lambda e: e.path)
        raise EidoValidationError(
            f"Validation unsuccessful. {len(errors)} errors found.", errors
        )


def validate_project(project, schema, exclude_case=False):
    """
    Validate a project object against a schema

    :param peppy.Sample project: a project object to validate
    :param str | dict schema: schema dict to validate against or a path to one
    :param bool exclude_case: whether to exclude validated objects
    from the error. Useful when used ith large projects
    """
    schema_dicts = read_schema(schema=schema)
    for schema_dict in schema_dicts:
        project_dict = project.to_dict()
        _validate_object(project_dict, preprocess_schema(schema_dict), exclude_case)
        _LOGGER.debug("Project validation successful")


def _validate_sample_object(sample, schemas, exclude_case=False):
    """
    Internal function that allows to validate a peppy.Sample object without
    requiring a reference to peppy.Project.

    :param peppy.Sample sample: a sample object to validate
    :param list[dict] schemas: list of schemas to validate against or a path to one
    :param bool exclude_case: whether to exclude validated objects
        from the error. Useful when used ith large projects
    """
    for schema_dict in schemas:
        schema_dict = preprocess_schema(schema_dict)
        sample_schema_dict = schema_dict[PROP_KEY]["_samples"]["items"]
        _validate_object(sample, sample_schema_dict, exclude_case)
        _LOGGER.debug(
            f"{getattr(sample, 'sample_name', '')} sample validation successful"
        )


def validate_sample(project, sample_name, schema, exclude_case=False):
    """
    Validate the selected sample object against a schema

    :param peppy.Project project: a project object to validate
    :param str | int sample_name: name or index of the sample to validate
    :param str | dict schema: schema dict to validate against or a path to one
    :param bool exclude_case: whether to exclude validated objects
        from the error. Useful when used ith large projects
    """
    sample = (
        project.samples[sample_name]
        if isinstance(sample_name, int)
        else project.get_sample(sample_name)
    )
    _validate_sample_object(
        sample=sample, schemas=read_schema(schema=schema), exclude_case=exclude_case
    )


def validate_config(project, schema, exclude_case=False):
    """
    Validate the config part of the Project object against a schema

    :param peppy.Project project: a project object to validate
    :param str | dict schema: schema dict to validate against or a path to one
    :param bool exclude_case: whether to exclude validated objects
        from the error. Useful when used ith large projects
    """
    schema_dicts = read_schema(schema=schema)
    for schema_dict in schema_dicts:
        schema_cpy = preprocess_schema(dpcpy(schema_dict))
        try:
            del schema_cpy[PROP_KEY]["_samples"]
        except KeyError:
            pass
        if "required" in schema_cpy:
            try:
                schema_cpy["required"].remove("_samples")
            except ValueError:
                pass
        project_dict = project.to_dict()
        _validate_object(project_dict, schema_cpy, exclude_case)
        _LOGGER.debug("Config validation successful")


def validate_inputs(sample, schema, exclude_case=False):
    """
    Determine which of this Sample's required attributes/files are missing
    and calculate sizes of the inputs.

    The names of the attributes that are required and/or deemed as inputs
    are sourced from the schema, more specifically from required_input_attrs
    and input_attrs sections in samples section. Note, this function does
    perform actual Sample object validation with jsonschema.

    :param peppy.Sample sample: sample to investigate
    :param list[dict] | str schema: schema dict to validate against or a path to one
    :return dict: dictionary with validation data, i.e missing,
        required_inputs, all_inputs, input_file_size
    :param bool exclude_case: whether to exclude validated objects
        from the error. Useful when used ith large projects
    :raise ValidationError: if any required sample attribute is missing
    """

    def _get_attr_values(obj, attrlist):
        """
        Get value corresponding to each given attribute.

        :param Mapping obj: an object to get the attributes from
        :param str | Iterable[str] attrlist: names of attributes to
            retrieve values for
        :return dict: value corresponding to
            each named attribute; null if this Sample's value for the
            attribute given by the argument to the "attrlist" parameter is
            empty/null, or if this Sample lacks the indicated attribute
        """
        # If attribute is None, then value is also None.
        if not attrlist:
            return None
        if not isinstance(attrlist, list):
            attrlist = [attrlist]
        # Strings contained here are appended later so shouldn't be null.
        return list(flatten([getattr(obj, attr, "") for attr in attrlist]))

    if isinstance(schema, str):
        schema = read_schema(schema)

    # validate attrs existence first
    _validate_sample_object(schemas=schema, sample=sample, exclude_case=exclude_case)

    all_inputs = set()
    required_inputs = set()
    schema = schema[-1]  # use only first schema, in case there are imports
    sample_schema_dict = schema["properties"]["_samples"]["items"]
    if FILES_KEY in sample_schema_dict:
        all_inputs.update(_get_attr_values(sample, sample_schema_dict[FILES_KEY]))
    if REQUIRED_FILES_KEY in sample_schema_dict:
        required_inputs = set(
            _get_attr_values(sample, sample_schema_dict[REQUIRED_FILES_KEY])
        )
        all_inputs.update(required_inputs)
    with cw(record=True) as w:
        input_file_size = sum(
            [size(f, size_str=False) or 0.0 for f in all_inputs if f != ""]
        ) / (1024**3)
        if w:
            _LOGGER.warning(
                f"{len(w)} input files missing, job input size was "
                f"not calculated accurately"
            )

    return {
        MISSING_KEY: [i for i in required_inputs if not os.path.exists(i)],
        REQUIRED_INPUTS_KEY: required_inputs,
        ALL_INPUTS_KEY: all_inputs,
        INPUT_FILE_SIZE_KEY: input_file_size,
    }



from logging import getLogger

from peppy.utils import load_yaml

from .const import *

_LOGGER = getLogger(__name__)


def preprocess_schema(schema_dict):
    """
    Preprocess schema before validation for user's convenience

    Preprocessing includes:
    - renaming 'samples' to '_samples' since in the peppy.Project object
        _samples attribute holds the list of peppy.Samples objects.
    - adding array of strings entry for every string specified to accommodate
        subsamples in peppy.Project

    :param dict schema_dict: schema dictionary to preprocess
    :return dict: preprocessed schema
    """
    _LOGGER.debug(f"schema ori: {schema_dict}")
    if "config" in schema_dict[PROP_KEY]:
        schema_dict[PROP_KEY]["_config"] = schema_dict[PROP_KEY]["config"]
        del schema_dict[PROP_KEY]["config"]
    else:
        _LOGGER.debug("No config section found in schema")
    if "samples" in schema_dict[PROP_KEY]:
        schema_dict[PROP_KEY]["_samples"] = schema_dict[PROP_KEY]["samples"]
        del schema_dict[PROP_KEY]["samples"]
        if "required" in schema_dict:
            schema_dict["required"][
                schema_dict["required"].index("samples")
            ] = "_samples"
        if (
            "items" in schema_dict[PROP_KEY]["_samples"]
            and PROP_KEY in schema_dict[PROP_KEY]["_samples"]["items"]
        ):
            s_props = schema_dict[PROP_KEY]["_samples"]["items"][PROP_KEY]
            for prop, val in s_props.items():
                if "type" in val and val["type"] in ["string", "number", "boolean"]:
                    s_props[prop] = {}
                    s_props[prop]["anyOf"] = [val, {"type": "array", "items": val}]
    else:
        _LOGGER.debug("No samples section found in schema")
    _LOGGER.debug(f"schema processed: {schema_dict}")
    return schema_dict


def read_schema(schema):
    """
    Safely read schema from YAML-formatted file.

    If the schema imports any other schemas, they will be read recursively.

    :param str | Mapping schema: path to the schema file
        or schema in a dict form
    :return list[dict]: read schemas
    :raise TypeError: if the schema arg is neither a Mapping nor a file path or
        if the 'imports' sections in any of the schemas is not a list
    """

    def _recursively_read_schemas(x, lst):
        if "imports" in x:
            if isinstance(x["imports"], list):
                for sch in x["imports"]:
                    lst.extend(read_schema(sch))
            else:
                raise TypeError("In schema the 'imports' section has to be a list")
        lst.append(x)
        return lst

    schema_list = []
    if isinstance(schema, str):
        return _recursively_read_schemas(load_yaml(schema), schema_list)
    elif isinstance(schema, dict):
        return _recursively_read_schemas(schema, schema_list)
    raise TypeError(
        f"schema has to be a dict, path to an existing file or URL to a remote one. "
        f"Got: {type(schema)}"
    )



from logging import CRITICAL, DEBUG, ERROR, INFO, WARN

from peppy.const import SAMPLE_NAME_ATTR
from peppy import __version__ as peppy_version
from ubiquerg import VersionInHelpParser

from . import __version__
from .const import *

LEVEL_BY_VERBOSITY = [
    ERROR,
    CRITICAL,
    WARN,
    INFO,
    DEBUG,
]

version_combined = f"{__version__} (peppy {peppy_version})"


def build_argparser():
    banner = "%(prog)s - Interact with PEPs"
    additional_description = "\nhttp://eido.databio.org/"

    parser = VersionInHelpParser(
        prog=PKG_NAME,
        description=banner,
        epilog=additional_description,
        version=version_combined,
    )

    subparsers = parser.add_subparsers(dest="command")
    parser.add_argument(
        "--verbosity",
        dest="verbosity",
        type=int,
        choices=range(len(LEVEL_BY_VERBOSITY)),
        help="Choose level of verbosity (default: %(default)s)",
    )
    parser.add_argument("--logging-level", dest="logging_level", help="logging level")
    parser.add_argument(
        "--dbg",
        dest="dbg",
        action="store_true",
        help="Turn on debug mode (default: %(default)s)",
    )
    sps = {}
    for cmd, desc in SUBPARSER_MSGS.items():
        sps[cmd] = subparsers.add_parser(cmd, description=desc, help=desc)
        sps[cmd].add_argument(
            "--st-index",
            required=False,
            type=str,
            help=f"Sample table index to use, samples are identified by '{SAMPLE_NAME_ATTR}' by default.",
        )
        if cmd != CONVERT_CMD:
            sps[cmd].add_argument(
                "pep",
                metavar="PEP",
                help="Path to a PEP configuration file in yaml format.",
                default=None,
            )
        else:
            sps[cmd].add_argument(
                "pep",
                metavar="PEP",
                nargs="?",
                help="Path to a PEP configuration file in yaml format.",
                default=None,
            )

    sps[VALIDATE_CMD].add_argument(
        "-s",
        "--schema",
        required=True,
        help="Path to a PEP schema file in yaml format.",
        metavar="S",
    )

    sps[VALIDATE_CMD].add_argument(
        "-e",
        "--exclude-case",
        default=False,
        action="store_true",
        help="Whether to exclude the validation case from an error. "
        "Only the human readable message explaining the error will "
        "be raised. Useful when validating large PEPs.",
    )

    sps[INSPECT_CMD].add_argument(
        "-n",
        "--sample-name",
        required=False,
        nargs="+",
        help="Name of the samples to inspect.",
        metavar="SN",
    )

    sps[INSPECT_CMD].add_argument(
        "-l",
        "--attr-limit",
        required=False,
        type=int,
        default=10,
        help="Number of sample attributes to display.",
    )

    group = sps[VALIDATE_CMD].add_mutually_exclusive_group()

    group.add_argument(
        "-n",
        "--sample-name",
        required=False,
        help="Name or index of the sample to validate. "
        "Only this sample will be validated.",
        metavar="S",
    )

    group.add_argument(
        "-c",
        "--just-config",
        required=False,
        action="store_true",
        default=False,
        help="Whether samples should be excluded from the validation.",
    )

    sps[CONVERT_CMD].add_argument(
        "-f",
        "--format",
        required=False,
        default="yaml",
        help="Output format (name of filter; use -l to see available).",
    )

    sps[CONVERT_CMD].add_argument(
        "-n",
        "--sample-name",
        required=False,
        nargs="+",
        help="Name of the samples to inspect.",
    )

    sps[CONVERT_CMD].add_argument(
        "-a",
        "--args",
        nargs="+",
        action="append",
        required=False,
        default=None,
        help="Provide arguments to the filter function (e.g. arg1=val1 arg2=val2).",
    )

    sps[CONVERT_CMD].add_argument(
        "-l",
        "--list",
        required=False,
        default=False,
        action="store_true",
        help="List available filters.",
    )

    sps[CONVERT_CMD].add_argument(
        "-d",
        "--describe",
        required=False,
        default=False,
        action="store_true",
        help="Show description for a given filter.",
    )

    sps[CONVERT_CMD].add_argument(
        "-p",
        "--paths",
        nargs="+",
        help="Paths to dump conversion result as key=value pairs.",
    )
    return parser, sps



import logging
import sys

from logmuse import init_logger
from peppy import Project

from .argparser import LEVEL_BY_VERBOSITY, build_argparser
from .const import *
from .conversion import (
    convert_project,
    get_available_pep_filters,
    pep_conversion_plugins,
)
from .exceptions import EidoFilterError
from .inspection import inspect_project
from .validation import validate_config, validate_project, validate_sample


def _parse_filter_args_str(input):
    """
    Parse user input specification.

    :param Iterable[Iterable[str]] input: user command line input,
        formatted as follows: [[arg=txt, arg1=txt]]
    :return dict: mapping of keys, which are input names and values
    """
    lst = []
    for i in input or []:
        lst.extend(i)
    return (
        {x.split("=")[0]: x.split("=")[1] for x in lst if "=" in x}
        if lst is not None
        else lst
    )


def main():
    """Primary workflow"""
    parser, sps = build_argparser()
    args, remaining_args = parser.parse_known_args()

    if args.command is None:
        parser.print_help(sys.stderr)
        sys.exit(1)

    # Set the logging level.
    if args.dbg:
        # Debug mode takes precedence and will listen for all messages.
        level = args.logging_level or logging.DEBUG
    elif args.verbosity is not None:
        # Verbosity-framed specification trumps logging_level.
        level = LEVEL_BY_VERBOSITY[args.verbosity]
    else:
        # Normally, we're not in debug mode, and there's not verbosity.
        level = LOGGING_LEVEL

    logger_kwargs = {"level": level, "devmode": args.dbg}
    init_logger(name="peppy", **logger_kwargs)
    global _LOGGER
    _LOGGER = init_logger(name=PKG_NAME, **logger_kwargs)

    if args.command == CONVERT_CMD:
        filters = get_available_pep_filters()
        if args.list:
            _LOGGER.info("Available filters:")
            if len(filters) < 1:
                _LOGGER.info("No available filters")
            for filter_name in filters:
                _LOGGER.info(f" - {filter_name}")
            sys.exit(0)
        if not "format" in args:
            _LOGGER.info("The following arguments are required: --format")
            sps[CONVERT_CMD].print_help(sys.stderr)
            sys.exit(1)
        if args.describe:
            if args.format not in filters:
                raise EidoFilterError(
                    f"'{args.format}' filter not found. Available filters: {', '.join(filters)}"
                )
            filter_functions_by_name = pep_conversion_plugins()
            print(filter_functions_by_name[args.format].__doc__)
            sys.exit(0)
        if args.pep is None:
            sps[CONVERT_CMD].print_help(sys.stderr)
            _LOGGER.info("The following arguments are required: PEP")
            sys.exit(1)
        if args.paths:
            paths = {y[0]: y[1] for y in [x.split("=") for x in args.paths]}
        else:
            paths = None

        p = Project(args.pep, sample_table_index=args.st_index)
        plugin_kwargs = _parse_filter_args_str(args.args)

        # append paths
        plugin_kwargs["paths"] = paths

        convert_project(p, args.format, plugin_kwargs)
        _LOGGER.info("Conversion successful")
        sys.exit(0)

    _LOGGER.debug(f"Creating a Project object from: {args.pep}")
    if args.command == VALIDATE_CMD:
        p = Project(cfg=args.pep, sample_table_index=args.st_index)
        if args.sample_name:
            try:
                args.sample_name = int(args.sample_name)
            except ValueError:
                pass
            _LOGGER.debug(
                f"Comparing Sample ('{args.pep}') in Project ('{args.pep}') "
                f"against a schema: {args.schema}"
            )
            validate_sample(p, args.sample_name, args.schema, args.exclude_case)
        elif args.just_config:
            _LOGGER.debug(
                f"Comparing Project ('{args.pep}') against a schema: {args.schema}"
            )
            validate_config(p, args.schema, args.exclude_case)
        else:
            _LOGGER.debug(
                f"Comparing Project ('{args.pep}') against a schema: {args.schema}"
            )
            validate_project(p, args.schema, args.exclude_case)
        _LOGGER.info("Validation successful")

    if args.command == INSPECT_CMD:
        p = Project(cfg=args.pep, sample_table_index=args.st_index)
        inspect_project(p, args.sample_name, args.attr_limit)
        sys.exit(0)



""" Exceptions for specific eido issues. """

from abc import ABCMeta

_all__ = ["PathAttrNotFoundError", "EidoSchemaInvalidError", "EidoFilterError"]


class EidoException(Exception):
    """Base type for custom package errors."""

    __metaclass__ = ABCMeta


class PathAttrNotFoundError(EidoException):
    """Path-like argument does not exist."""

    def __init__(self, key):
        super(PathAttrNotFoundError, self).__init__(key)


class EidoSchemaInvalidError(EidoException):
    """Schema does not comply to eido-specific requirements."""

    def __init__(self, key):
        super(EidoSchemaInvalidError, self).__init__(key)


class EidoFilterError(EidoException):
    """Issue with the PEP filter."""

    def __init__(self, key):
        super(EidoFilterError, self).__init__(key)


class EidoValidationError(EidoException):
    """Object was not validated successfully according to schema."""

    def __init__(self, message, errors):
        super().__init__(message)
        self.errors = errors
